{ Unit: ThreadTimer
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    This is like normal TTimer component expect this uses separate thread and
    Sleep() for the timing. This should be pretty accurate timer, atleast
    much better than TTimer..

    WARNING: Remember to set Synchronized to True if your OnTimer event
      updates visual VCL elements. Read more from TThread.Synchronize() help.
      If you don't need synchronization then set it false because it makes the
      thread faster -> more accurate timer!   For "safety" reasons it's set
      True by default.

    You can also fiddle with Priority, but be careful if you're executing
    something which eats CPU time..

    Notes about thread: Thread is created when Enabled is set true (e.g. if
    it's set True in design-time thread is created in RUN-TIME during form
    creation. So, no threads in design-time. If timer is destroyed or Enabled
    is set to False, thread is destroyed (freed on terminate). So, new thread everytime
    when timer is enabled.

    OnTimer is called first time after first Interval pause (just like in TTimer).

    New properties:

      Priority: TThreadPriority    // thread's priority
      Synchronize: Boolean         // if this is true then OnTimer event is
                                      synchronized
      WaitForTermination: Boolean  // waits for the thread to terminate when Enabled is
                                   // set true. note, with big intervals it might hang your
                                   // app for the interval time.
                                   // Never the less, this is useful if you need to
                                   // free objects which the thread uses. Optionally
                                   // make the thread to check objects before use

      There is also public read-only property,
      Thread: TExecuteThread       // gives direct access to thread, but be
                                   // careful, you may confuse the parent Timer.
                                   // Thread can point to nil. So check before
                                   // using.

  History:

    2001-10-11: Changed to TBcThreadTimer and added to Bluecave Components
                package
    2001-03-27: Added WaitForTermination property
    2001-01-24: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcThreadTimer;

interface

uses
  Windows, Messages, SysUtils, Classes;

type
  TBcThreadTimer = class;

  TExecuteThread = class(TThread)
  private
    FOwner: TBcThreadTimer;
    FInterval: Integer;
    FSynchronized: Boolean;
  public
    constructor Create(AOwner: TBcThreadTimer; CreateSuspended: Boolean);
    procedure RunEvent;
    procedure Execute; override;
    property Interval: Integer read FInterval write FInterval;
    property Synchronized: Boolean read FSynchronized write FSynchronized;
  end;

  TBcThreadTimer = class(TComponent)
  private
    { Private declarations }
    FOnTimer: TNotifyEvent;
    FEnabled: Boolean;
    FInterval: Integer;
    FThread: TExecuteThread;
    FSynchronized: Boolean;
    FPriority: TThreadPriority;
    FWaitForTermination: Boolean;
    function GetEnabled: Boolean;
    procedure SetEnabled(const Value: Boolean);
    function GetSynchronized: Boolean;
    procedure SetSynchronized(const Value: Boolean);
    function GetInterval: Integer;
    procedure SetInterval(const Value: Integer);
    function GetPriority: TThreadPriority;
    procedure SetPriority(const Value: TThreadPriority);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Thread: TExecuteThread read FThread; { read-only }
  published
    { Published declarations }
    property Enabled: Boolean read GetEnabled write SetEnabled default False;
    property WaitForTermination: Boolean read FWaitForTermination write FWaitForTermination default False;
    property Synchronized: Boolean read GetSynchronized write SetSynchronized default True;
    property Interval: Integer read GetInterval write SetInterval default 1000;
    property Priority: TThreadPriority read GetPriority write SetPriority default tpNormal;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
  end;

implementation

constructor TExecuteThread.Create(AOwner: TBcThreadTimer; CreateSuspended: Boolean);
begin
  TThread(Self).Create(CreateSuspended);
  FOwner := AOwner;
end;

procedure TExecuteThread.RunEvent;
begin
  FOwner.FOnTimer(FOwner); // We can execute Owner's private procedure inside
                           // same unit
end;

procedure TExecuteThread.Execute;
begin
  repeat
    if Terminated then Break;
    if FInterval > 0 then Sleep(FInterval);
    if Terminated then Break;
    if Synchronized then
      Synchronize(RunEvent)
    else
      RunEvent;
  until Terminated;
end;

constructor TBcThreadTimer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FSynchronized := True;
  FInterval := 1000;
  FPriority := tpNormal;
  FEnabled := False;
  FWaitForTermination := False;
end;

destructor TBcThreadTimer.Destroy;
begin
  if Assigned(FThread) then FThread.Terminate;
  inherited Destroy;
end;

function TBcThreadTimer.GetSynchronized: Boolean;
begin
  Result := FSynchronized;
end;

procedure TBcThreadTimer.SetSynchronized(const Value: Boolean);
begin
  if FSynchronized = Value then Exit;
  FSynchronized := Value;
  if Assigned(FThread) and not (csDesigning in ComponentState) then
    FThread.Synchronized := FSynchronized;
end;

function TBcThreadTimer.GetEnabled: Boolean;
begin
{  Result := (Assigned(FThread) and (not FThread.Suspended));}
  Result := FEnabled;
end;

procedure TBcThreadTimer.SetEnabled(const Value: Boolean);
begin
  if FEnabled = Value then Exit;

  FEnabled := Value;
  if not (csDesigning in ComponentState) then
  begin
    if FEnabled then
    begin
      if Assigned(FThread) then FThread.Terminate;
      FThread := TExecuteThread.Create(Self, True);
      FThread.FreeOnTerminate := True;
      FThread.Priority := FPriority;
      FThread.Interval := FInterval;
      FThread.Resume;
    end
    else if Assigned(FThread) then
    begin
      if FWaitForTermination then
      begin
        FThread.FreeOnTerminate := False;
        FThread.Terminate;
        FThread.WaitFor;
        FreeAndNil(FThread);
      end
      else
        FThread.Terminate;

    end

  end;
end;

function TBcThreadTimer.GetInterval: Integer;
begin
  Result := FInterval;
end;

procedure TBcThreadTimer.SetInterval(const Value: Integer);
begin
  if FInterval = Value then Exit;
  FInterval := Value;
  if FInterval < 0 then FInterval := 0;
  if Assigned(FThread) and (not (csDesigning in ComponentState)) then
    FThread.Interval := FInterval;
end;

function TBcThreadTimer.GetPriority: TThreadPriority;
begin
  Result := FPriority;
end;

procedure TBcThreadTimer.SetPriority(const Value: TThreadPriority);
begin
  if FPriority = Value then Exit;
  FPriority := Value;
  if Assigned(FThread) and (not (csDesigning in ComponentState)) then
    FThread.Priority := FPriority;
end;

end.
